/* alea jacta est
 *
 * Parse a dice description and roll 'em.
 * Copyright (c) 2005 Amelia A. Lewis
 *
 * See COPYING for license.  This work is released under the GPL,
 * according to the prophecy.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/param.h>
#include <math.h>

#include "instloc.h"

char *progname;
char dice_pattern[MAXPATHLEN];

int ndice;
int nsides;
int modifier;

int zed_flag = 0;
int debug_flag = 0;

/*
 * usage
 *	Prints a usage statement.  Exits with return 1.
 */
void
usage(char *calledas)
{
    fprintf(stderr, "Usage:\n\t%s [-hvz] [[nD]n[+|-n]]\
\n\
\n      -h  Show this usage description and exit\
\n      -v  Show version and exit\
\n      -z  Index from zero (default 1) to n - 1 (default n)\
\n\
\n      Where n is a number of dice or number of sides (positive integer),\
\n        and D is the letter 'D' or 'd'\
\n", calledas);
    exit(1);
}

void
version(char *calledas)
{
    fprintf(stderr, "%s version %s\n", calledas, VERSION);
    exit(1);
}

int
toss_dice(int upper_limit)
{
    return (int) ((upper_limit * 1.0) * random() / (RAND_MAX + 1.0));
}

/*
 * getargs
 *	Parses the command line, setting flags.
 */
void
getargs(int ac, char *av[])
{
    int c;
    char deft[MAXPATHLEN];
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);
    
    if (!strcmp(progname, TIME_BIN))
        strcpy(deft, TIME_PATTERN_DEFAULT);
    else
        strcpy(deft, DICE_PATTERN_DEFAULT);

    while ((c = getopt(ac, av, "Dhvz")) != EOF)
        switch (c)
        {
            case 'D':
                debug_flag = 1;
                break;
            case 'v':
                version(progname);
                break;
            case 'z':
                zed_flag = 1;
                break;
            case 'h':
            case '?':
            default:
                usage(progname);
                break;
        }

    av += optind;
    if (*av)
        strcpy(dice_pattern, *av);
    else
        strcpy(dice_pattern, deft);
    
    if (strchr(dice_pattern, 'D'))
    {
        if (strchr(dice_pattern, '+'))
            sscanf(dice_pattern, "%dD%d+%d", &ndice, &nsides, &modifier);
        else if (strchr(dice_pattern, '-'))
            sscanf(dice_pattern, "%dD%d%d", &ndice, &nsides, &modifier);
        else
            sscanf(dice_pattern, "%dD%d", &ndice, &nsides);
    }
    else if (strchr(dice_pattern, 'd'))
    {
        if (strchr(dice_pattern, '+'))
            sscanf(dice_pattern, "%dd%d+%d", &ndice, &nsides, &modifier);
        else if (strchr(dice_pattern, '-'))
            sscanf(dice_pattern, "%dd%d%d", &ndice, &nsides, &modifier);
        else
            sscanf(dice_pattern, "%dd%d", &ndice, &nsides);
    }
    else
    {
        ndice = 1;
        if (strchr(dice_pattern, '+'))
            sscanf(dice_pattern, "%d+%d", &nsides, &modifier);
        else if (strchr(dice_pattern, '-'))
            sscanf(dice_pattern, "%d%d", &nsides, &modifier);
        else
            sscanf(dice_pattern, "%d", &nsides);
    }

    if ( (ndice < 1) || (nsides < 1) )
        usage(progname);

}

/*
 * main
 *	If you need a comment for main, I suggest you try K&R
 */
int
main(int argc, char *argv[])
{
    int i, total = 0, die = 0;

    srandom(time(NULL) + getpid());
    getargs(argc, argv);
    
    for (i = 0; i < ndice; i++)
    {
        die = toss_dice(nsides) + (zed_flag ? 0 : 1);
        if (debug_flag)
            fprintf(stderr, "%d ", die);
        total += die;
    }
    if (modifier && debug_flag)
        fprintf(stderr, "+%d = %d\n", modifier, total);
    total += modifier;
    if (modifier)
        fprintf(stdout, "%dD%d%+d : %d\n", ndice, nsides, modifier, total);
    else
        fprintf(stdout, "%dD%d : %d\n", ndice, nsides, total);
    
    return 0;
}
