/* bmi
 *
 * given two of: height, weight, bmi, calculate the third.
 *
 * Copyright (c) 2005 Amelia A. Lewis
 *
 * See COPYING for license.  This work is released under the GPL,
 * according to the prophecy.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/param.h>
#include <math.h>

#include "instloc.h"

char *progname;

int height = 0;
int weight = 0;
int bmi = 0;

int solution_flag = 27;

/*
 * usage
 *	Prints a usage statement.  Exits with return 1.
 */
void
usage(char *calledas)
{
    fprintf(stderr, "Usage:\n\t%s [-b bmi][-h height][-w weight][-uv]\\
\n\
\n      -b  bmi: should be decimal (21.0)\
\n      -h  height: in inches\
\n      -u  Show this usage statement\
\n      -v  Show version\
\n      -w  weight: in pounds\
\n\
\n", calledas);
    exit(1);
}

void
version(char *calledas)
{
    fprintf(stderr, "%s version %s\n", calledas, VERSION);
    exit(1);
}

/*
 * getargs
 *	Parses the command line, setting flags.
 */
void
getargs(int ac, char *av[])
{
    int c;
    double b;
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);

    while ((c = getopt(ac, av, "b:h:uvw:")) != EOF)
        switch (c)
        {
            case 'b':
                if (*optarg)
                {
                    sscanf(optarg, "%lf", &b);
                    bmi = (int)(10.0 * b);
                }
                break;
            case 'h':
                if (*optarg)
                    height = (int)strtoul(optarg, NULL, 10);
                break;
            case 'v':
                version(progname);
                break;
            case 'w':
                if (*optarg)
                    weight = (int)strtoul(optarg, NULL, 10);
                break;
            case 'u':
            case '?':
            default:
                usage(progname);
                break;
        }

    if (!bmi)
    {
        if (!weight || !height)
            usage(progname);
        solution_flag = 0;
    }
    if (!height)
    {
        if (!bmi || !weight)
            usage(progname);
        solution_flag = 1;
    }
    if (!weight)
    {
        if (!height || !bmi)
            usage(progname);
        solution_flag = -1;
    }
}

int
solve_height()
{
    return (int)sqrt(703.0 * weight / (bmi / 10.0));
}

int
solve_bmi()
{
    return (int)(7030.0 * weight / height / height);
}

int
solve_weight()
{
    return (int)(1.0 * height * height * (bmi / 10.0) / 703.0);
}


/*
 * main
 *	If you need a comment for main, I suggest you try K&R
 */
int
main(int argc, char *argv[])
{
    getargs(argc, argv);
    
    switch (solution_flag)
    {
        case -1:
            weight = solve_weight();
            break;
        case 0:
            bmi = solve_bmi();
            break;
        case 1:
            height = solve_height();
            break;
        default:
            usage(progname);
    }
    
    fprintf(stdout, "%d'%d\", %dlbs. (BMI: %3.1f)\n", height / 12, height % 12, weight, bmi / 10.0);

    return 0;
}
