/* Extras
 *
 * "extras" allows generation of that staple of fiction, the walkon/walkoff
 * extra, with a certain amount of pleasing detail.  It is mostly suitable
 * for fiction set in modern (late twentieth-century America).
 *
 * extras started as a name generator.  It should include birthdates,
 * appearance, and random silliness by the time that you see it.
 *
 * Copyright (c) 2005 Amelia A. Lewis
 *
 * See COPYING for license.  This work is released under the GPL,
 * according to the prophecy.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/param.h>
#include <math.h>

#include "instloc.h"

char given_fn[MAXPATHLEN], sur_fn[MAXPATHLEN], mbti_fn[MAXPATHLEN];
char height_fn[MAXPATHLEN], bmi_fn[MAXPATHLEN], color_fn[MAXPATHLEN];
char *progname;
FILE *givenf, *surf, *mbtif, *heightf, *bmif, *colorf;

char *hair_length[] = { "very long", "long", "medium", "short", "very short", "radical" };
char *hair_texture[] = { "nappy", "curly", "wavy", "straight" };

char *mbti_index[] = { "ISTJ", "ISFJ", "INFJ", "INTJ", "ISTP", "ISFP", "INFP", "INTP",
                       "ESTJ", "ESFJ", "ENFJ", "ENTJ", "ESTP", "ESFP", "ENFP", "ENTP" };
char *mbti_name[] = { "dutiful", "nurturer", "protector", "scientist",
                      "mechanic", "artist", "idealist", "thinker",
                      "guardian", "caregiver", "giver", "executive",
                      "visionary", "doer", "inspirer", "performer" };

char *mbti_bits_index[] = { "I", "E", "S", "N", "T", "F", "P", "J" };
char *mbti_bits[] = { "introverted", "extroverted", "sensing", "intuitive", "thinking", "feeling", "perceptive", "judging" };

/* www.typelogic.com/lc.html */

int babble_flag = 0;
int lastname_flag = 1;
int poison_flag = -1;
int firstname_flag = 0;

int personality_flag = 0;
int size_flag = 0;
int birth_flag = 0;
int year_flag = 0;
int quirk_flag = 0;
int looks_flag = 0;
int menses_flag = 0;

int prophecy_flag = 0;
int salute_flag = 0;

int base_year = BASE_YEAR;

/*
 * usage
 *	Prints a usage statement.  Exits with return 1.
 */
void
usage(char *calledas)
{
    fprintf(stderr, "Usage:\n\t%s [-aAbdfFhlmNnprsvxy] [-Y year]\
\n\
\n      -A  Generate \"all\" information\
\n      -a  Include \"appearance\" (eyes, hair, skin)\
\n      -b  Include birthdate (month, day, hour, minute)\
\n      -d  Include a random disadvantage/quirk (not impl)\
\n      -f  Set gender female\
\n      -F  With -f or -m, generate three-name \"full\" name\
\n      -h  Show this usage description\
\n      -l  Last name only\
\n      -M  For women, generate menstrual cycle info\
\n      -m  Set gender male\
\n      -N  Do NOT generate name (at all)\
\n      -n  First name only\
\n      -p  Include personality (Myers/Briggs)\
\n      -r  Have random sex\
\n      -s  Include \"size\" information (height, weight, BMI)\
\n      -v  Show version\
\n      -x  Explain things\
\n      -y  Include year of birth\
\n      -Y year  Set base year to year, include random year of birth\
\n\
\nWith no options, prints last name or random-sex full name (compile-time option)\
\n", calledas);
    exit(1);
}

void
explain(char *calledas)
{
    fprintf(stderr, "%s v. %s: a program for generating information about minor characters.\
\n\
\n\tActually, a suite of programs: extras, goon, bimbo, and pains.\
\n\tThese names may be changed at compile time.\
\n\textras: base program\
\n\tgoon: equivalent to extras -mn\
\n\tbimbo: equivalent to extras -fn\
\n\tpains: generate height, weight, and BMI given sex + adult height+BMI\
\n\tmbti: doesn't exist yet\
\n\
\n\tNames are generated randomly based on frequency information from\
\n\tthe census.\
\n\
\n\t[UNUSUAL] as a name means something not in the lists, usually odd\
\n\tspelling, unusual ethnicity, or just STRANGE.\
\n\tIf you can't think of something appropriate, run extras (or goon\
\n\tor bimbo, \n\tdepending) and ellis island the result.\
\n\
\n\tHeight, weight, BMI, first and middle names, and MBTI are\
\n\tsex-dependent, so can't be generated without specifying sex.\
\n\tOther variables (last name, appearance, disadvantages, birthdate,\
\n\tbirth year, and so on) are not sex-dependent.\
\n\tMenstrual information is available only when -f is specified.\
\n\
\n\tHeight is from CDC data per age per sex; so is BMI.  Weight is\
\n\tgenerated from height and BMI.  See pains for more information.\
\n\
\n\tSkin color is roughly correlated with Census \"race\" indicators.\
\n\tSkin, hair, and eye color are correlated.  Hair length and texture\
\n\tis pretty much random.\
\n\
\n\tDisadvantages/quirks aren't implemented yet.  *sigh*\
\n\
\n\tMBTI profile is randomly selected based on weighting in the population\
\n\tfor each sex.  For more information, see the (nonexistent) mbti program.\
\n\
\n\tDiscomfort index: scale of 0 (none) to 5 (debilitating)\
\n\t 0: none, light flow; 1: average flow max; 2: easy, average flow\
\n\t 3: average flow, some cramps; 4: heavy flow, painful cramps, bloating\
\n\t 5: debilitating (very heavy flow, bad cramps, distracting and awful)\
\n\
\n\t\"In accordance with the prophecy.\" Something weird happened.\
\n\
\n", calledas, VERSION);
    exit(1);
}

void
version(char *calledas)
{
    fprintf(stderr, "%s version %s\n", calledas, VERSION);
    exit(1);
}

/* for debugging only */
void
print_flags(char *proggie)
{
    fprintf(stderr, "Called as: %s\n", proggie);
    fprintf(stderr, "First Name: %d; ", firstname_flag);
    fprintf(stderr, "Last Name: %d; ", lastname_flag);
    fprintf(stderr, "Poison: %d\n", poison_flag);
    fprintf(stderr, "Appearance: %d; ", looks_flag);
    fprintf(stderr, "Size: %d\n", size_flag);
    fprintf(stderr, "Birthdate: %d; ", birth_flag);
    fprintf(stderr, "Year: %d\n", year_flag);
    fprintf(stderr, "Personality: %d; ", personality_flag);
    fprintf(stderr, "Quirks: %d\n", quirk_flag);
    fprintf(stderr, "Babble: %d; ", babble_flag);
    fprintf(stderr, "Prophecy: %d\n", prophecy_flag);
}

int
toss_dice(int upper_limit)
{
    return (1 + (int) ((upper_limit * 1.0) * random() / (RAND_MAX + 1.0)));
}

/*
 * getargs
 *	Parses the command line, setting flags.
 */
void
getargs(int ac, char *av[])
{
    int c;
    int name_override_flag = 0;
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);

    if (!strcmp(progname, FEMALE_BIN))
    {
        poison_flag = 0;
        firstname_flag = 1;
        lastname_flag = 0;
    }
    else if (!strcmp(progname, MALE_BIN))
    {
        poison_flag = 1;
        firstname_flag = 1;
        lastname_flag = 0;
    }

    if (FULLNAME_DEFAULT)
    {
        firstname_flag = 1;
        lastname_flag = 1;
    }

    while ((c = getopt(ac, av, "AabdFfhlMmNnprsvxY:y")) != EOF)
        switch (c)
        {
            case 'a':
                looks_flag = 1;
                break;
            case 'A':
                looks_flag = 1;
                firstname_flag = 1;
                lastname_flag = 1;
                personality_flag = 1;
                birth_flag = 1;
/*                quirk_flag = 1; */
                size_flag = 1;
                if (FULL_BIRTHDATE)
                    year_flag = 1;
                if (!poison_flag)
                    menses_flag = 1;
                break;
            case 'b':
                birth_flag = 1;
                if (FULL_BIRTHDATE)
                    year_flag = 1;
                break;
            case 'd':
                quirk_flag = 1;
                break;
            case 'f':
                poison_flag = 0;
                break;
            case 'F':
                firstname_flag = 1;
                lastname_flag = 1;
                break;
            case 'l':
                firstname_flag = 0;
                lastname_flag = 1;
                if (!name_override_flag)
                    name_override_flag = 1;
                else
                    name_override_flag = -1;
                break;
            case 'M':
                menses_flag = 1;
                break;
            case 'm':
                poison_flag = 1;
                break;
            case 'N':
                firstname_flag = 0;
                lastname_flag = 0;
                if (!name_override_flag)
                    name_override_flag = 1;
                else
                    name_override_flag = -1;
                break;
            case 'n':
                firstname_flag = 1;
                lastname_flag = 0;
                if (!name_override_flag)
                    name_override_flag = 1;
                else
                    name_override_flag = -1;
                break;
            case 'p':
                personality_flag = 1;
                break;
            case 'r':
                if (poison_flag == -1)
                    poison_flag = toss_dice(2) - 1;
                break;
            case 's':
                size_flag = 1;
                break;
            case 'v':
                version(progname);
                break;
            case 'x':
                explain(progname);
                break;
            case 'y':
                year_flag = 1;
                break;
            case 'Y':
                year_flag = 1;
                if (*optarg)
                     base_year = (int)strtoul(optarg, NULL, 10);
                break;
            case 'h':
            case '?':
            default:
                usage(progname);
                break;
        }

    if (name_override_flag < 0)
    {
        fprintf(stderr, "%s: Invalid command line options (choose one of -lNn)\n", progname);
        usage(progname);
    }

    /* after parsing; set only if unset */
    if (SEX_IS_RANDOM && (poison_flag == -1))
    {
        poison_flag = toss_dice(2) - 1;
    }

    if (poison_flag != -1)
    {
        if (firstname_flag)
            sprintf(given_fn, "%s/%s", INSTALL_DATA, poison_flag ? "givenm" : "givenf");
        if (personality_flag)
            sprintf(mbti_fn, "%s/%s", INSTALL_DATA, poison_flag ? "mbtim" : "mbtif");
        if (size_flag)
        {
            sprintf(height_fn, "%s/%s", INSTALL_DATA, poison_flag ? "heightm" : "heightf");
            sprintf(bmi_fn, "%s/%s", INSTALL_DATA, poison_flag ? "bmim" : "bmif");
        }
        if (menses_flag && poison_flag)
        {
            fprintf(stderr, "%s: no menstrual cycle.  *THWACK*\n", progname);
            menses_flag = 0;
        }
    }
    else
    {
        if (size_flag)
            fprintf(stderr, "%s: cannot generate height/weight without gender.\n", progname);
        size_flag = 0; /* can't generate any reasonable size */
        if (personality_flag)
            fprintf(stderr, "%s: cannot generate MBTI profile without gender.\n", progname);
        personality_flag = 0; /* only have separate data files */
        if (firstname_flag)
           fprintf(stderr, "%s: cannot generate first name without gender.\n", progname);
        firstname_flag = 0;
    }
    if (lastname_flag)
        sprintf(sur_fn, "%s/%s", INSTALL_DATA, "sur");
    if (looks_flag)
    {
        sprintf(color_fn, "%s/%s", INSTALL_DATA, "color");
    }
/*    print_flags(progname); */
}

/*
 * getfiles
 *	Opens the files.
 */
void
getfiles(void)
{
    if (lastname_flag)
    {
        surf = fopen(sur_fn, "r");
        if (!(surf))
        {
            perror(sur_fn);
            exit(2);
        }
    }
    if (poison_flag != -1)
    {
        if (firstname_flag)
        {
            givenf = fopen(given_fn, "r");
            if (!(givenf))
            {
                perror(given_fn);
                exit(2);
            }
        }
        if (personality_flag)
        {
            mbtif = fopen(mbti_fn, "r");
            if (!(mbtif))
            {
                perror(mbti_fn);
                exit(2);
            }
        }
        if (size_flag)
        {
            heightf = fopen(height_fn, "r");
            if (!(heightf))
            {
                perror(height_fn);
                exit(2);
            }
            bmif = fopen(bmi_fn, "r");
            if (!(bmif))
            {
                perror(bmi_fn);
                exit(2);
            }
        }
    }
    if (looks_flag)
    {
        colorf = fopen(color_fn, "r");
        if (!(colorf))
        {
            perror(color_fn);
            exit(2);
        }
    }
}

/*
 * get_next_record
 *	Reads the next record, allocating and initializing an sdb_recptr
 * structure.
 */
int
get_next_record(FILE *dataf, char *n)
{
    int rval;
    
    if (fscanf(dataf, "%s %d", n, &rval) == EOF)
        exit(3);
    return rval;
}

int
get_next_record2(FILE *dataf, int *i)
{
    int rval;
    
    if (fscanf(dataf, "%d %d", i, &rval) == EOF)
        exit(3);
    return rval;
}

int
get_next_record3(FILE *dataf, char *eyes, char *hair, char *skin)
{
    int rval;
    
    if (fscanf(dataf, "%s %s %s %d", eyes, hair, skin, &rval) == EOF)
        exit(3);
    return rval;
}

int
toss_multiple(int dice, int sides, int per_die_mod)
{
    int acc = 0, i;
    
    for (i = 0; i < dice; i++)
        acc += toss_dice(sides) + per_die_mod;
    return acc;
}

/*
 * main
 *	If you need a comment for main, I suggest you try K&R
 */
int
main(int argc, char *argv[])
{
    int die = 0;
    int current;
    char name[16];
    strcpy(name, "");
    
    srandom(time(NULL) + getpid());
    getargs(argc, argv);
    getfiles();

    if ((poison_flag != -1) && (firstname_flag))
    {
        /* get a first name */
        die = toss_dice(NAMES_MAX);
        do
        {
            current = get_next_record(givenf, name);
        } while (die > current);
        /*if (!strcmp(name, UNIQUE)) prophecy_flag++;*/
        fprintf(stdout, "%s ", name);
        /* if fullname_flag is set, do it again (middle name) */
        if (lastname_flag)
        {
            rewind(givenf);
            die = toss_dice(NAMES_MAX);
            do
            {
                current = get_next_record(givenf, name);
            } while (die > current);
            /*if (!strcmp(name, UNIQUE)) prophecy_flag++;*/
            fprintf(stdout, "%s ", name);
        }
        else
            fprintf(stdout, "\n");
    }
    if (lastname_flag)
    {
        die = toss_dice(NAMES_MAX);
        /* get a last name */
        do
        {
            current = get_next_record(surf, name);
        } while (die > current);
        /*if (!strcmp(name, UNIQUE)) prophecy_flag++;*/
        fprintf(stdout, "%s\n", name);
    }
    if (birth_flag)
    {
        char zodiac = 0;
        int leap_flag = 0;
        die = toss_dice(24) - 1;
        fprintf(stdout, "Born %02d%c", die, ':');
        die = toss_dice(60) - 1;
        fprintf(stdout, "%02d ", die);
        die = toss_dice(BIRTH_MAX);
        /* decompose into month and day */
        /* uses gcc extension for case ranges */
        switch (die)
        {
            case 1 ... 124:
                fprintf(stdout, "%d Jan", die % 31 + 1);
                if (die % 31 + 1 < 21) zodiac = 10;
                else zodiac = 11;
                break;
            case 125 ... 237:
                if (die != 237)
                {
                    fprintf(stdout, "%d Feb", (die - 124) % 28 + 1);
                    if ((die - 124) % 28 + 1 < 21) zodiac = 11;
                    else zodiac = 12;
                }
                else
                {
                    prophecy_flag++;
                    leap_flag++;
                    fprintf(stdout, "29 Feb (Leap Day)");
                    zodiac = 12;
                }
                break;
            case 238 ... 361:
                fprintf(stdout, "%d Mar", (die - 237) % 31 + 1);
                if ((die - 237) % 31 + 1 < 21) zodiac = 12;
                else zodiac = 1;
                if ((die - 237) % 31 + 1 == 19) salute_flag++;
                break;
            case 362 ... 481:
                fprintf(stdout, "%d Apr", (die - 361) % 30 + 1);
                if ((die - 361) % 30 + 1 < 21) zodiac = 1;
                else zodiac = 2;
                break;
            case 482 ... 605:
                fprintf(stdout, "%d May", (die - 481) % 31 + 1);
                if ((die - 481) % 31 + 1 < 22) zodiac = 2;
                else zodiac = 3;
                break;
            case 606 ... 725:
                fprintf(stdout, "%d Jun", (die - 605) % 30 + 1);
                if ((die - 605) % 30 + 1 < 22) zodiac = 3;
                else zodiac = 4;
                break;
            case 726 ... 849:
                fprintf(stdout, "%d Jul", (die - 725) % 31 + 1);
                if ((die - 725) % 31 + 1 < 23) zodiac = 4;
                else zodiac = 5;
                break;
            case 850 ... 973:
                fprintf(stdout, "%d Aug", (die - 849) % 31 + 1);
                if ((die - 849) % 31 + 1 < 23) zodiac = 5;
                else zodiac = 6;
                break;
            case 974 ... 1093:
                fprintf(stdout, "%d Sep", (die - 973) % 30 + 1);
                if ((die - 974) % 30 + 1 < 24) zodiac = 6;
                else zodiac = 7;
                break;
            case 1094 ... 1217:
                fprintf(stdout, "%d Oct", (die - 1093) % 31 + 1);
                if ((die - 1093) % 31 + 1 < 24) zodiac = 7;
                else zodiac = 8;
                break;
            case 1218 ... 1337:
                fprintf(stdout, "%d Nov", (die - 1217) % 30 + 1);
                if (die == 1337) prophecy_flag++;
                if ((die - 1217) % 30 + 1 < 23) zodiac = 8;
                else zodiac = 9;
                break;
            case 1338 ... 1461:
                fprintf(stdout, "%d Dec", (die - 1337) % 31 + 1);
                if ((die - 1337) % 31 + 1 < 23) zodiac = 9;
                else zodiac = 10;
                break;
        }
        if (year_flag)
        {
            if (leap_flag)
            {
                do
                {
                    die = toss_dice(100);
                } while ((die + BASE_YEAR) % 4);
            }
            else
                die = toss_dice(100);
            fprintf(stdout, ", %d", die + base_year);
        }
        switch (zodiac)
        {
            case 1:
                fprintf(stdout, " (Aries)");
                break;
            case 2:
                fprintf(stdout, " (Taurus)");
                break;
            case 3:
                fprintf(stdout, " (Gemini)");
                break;
            case 4:
                fprintf(stdout, " (Cancer)");
                break;
            case 5:
                fprintf(stdout, " (Leo)");
                break;
            case 6:
                fprintf(stdout, " (Virgo)");
                break;
            case 7:
                fprintf(stdout, " (Libra)");
                break;
            case 8:
                fprintf(stdout, " (Scorpio)");
                break;
            case 9:
                fprintf(stdout, " (Sagittarius)");
                break;
            case 10:
                fprintf(stdout, " (Capricorn)");
                break;
            case 11:
                fprintf(stdout, " (Aquarius)");
                break;
            case 12:
                fprintf(stdout, " (Pisces)");
                break;
        }
        fprintf(stdout, "\n");
    }
    else if (year_flag)
    {
            die = toss_dice(100);
            fprintf(stdout, "Born %d\n", die + base_year);
    }
    if (size_flag)
    {
        int h, b;
        double height, bmi, weight;
        /* get the height */
        die = toss_dice(NAMES_MAX);
        do
        {
            current = get_next_record2(heightf, &h);
        } while (die > current);
        height = 1.0 * h;
        
        /* get bmi */
        die = toss_dice(NAMES_MAX);
        do
        {
            current = get_next_record2(bmif, &b);
        } while (die > current);
        bmi = b / 10.0;
        
        weight = bmi * height * height / 703.0;
        weight = rint(weight);
        
        fprintf(stdout, "%d'%d\", %d lbs. (BMI: %3.1f)\n", (int)(height / 12),
                (int) ((int)height % 12), (int) weight, bmi);
    }
    if (looks_flag)
    {
        char h[16], e[16], s[16];
        char *l, *t;
        
        die = toss_dice(NAMES_MAX);
        do
        {
            current = get_next_record3(colorf, e, h, s);
        } while (die > current);
        l = hair_length[toss_dice(6) - 1];
        t = hair_texture[toss_dice(4) - 1];
        fprintf(stdout, "%s skin, %s %s %s hair and %s eyes\n", s, l, t, h, e);
    }
    if (quirk_flag)
    {
/* not going to work with current get_next_record
        char descr[MAX_LINE_LEN];
        die = toss_dice(NAMES_MAX);
        do
        {
            current = get_next_string(quirksf, descr);
        } while (die > current);
        fprintf(stdout, "%s\n", descr); */
        fprintf(stderr, "Disadvantages stuff not implemented.\n");
    }
    if (personality_flag)
    {
        int index;
        die = toss_dice(NAMES_MAX);
        do
        {
            current = get_next_record(mbtif, name);
        } while (die > current);
        for (index = 0; index < 16; index++)
        {
            if (!strcmp(mbti_index[index], name))
                break;
        }
        fprintf(stdout, "MBTI profile: %s (%s)\n", name, mbti_name[index]);
    }
    if (menses_flag)
    {
        int cycle_length, cycle_var, duration, duration_var, last_onset;
        int discomfort_index_low, discomfort_index_high;
        cycle_length = toss_multiple(3, 4 * 24, -1) + 23 * 24; /* hours */
        cycle_var = toss_multiple(3, 24, 0);
        duration = toss_multiple(3, 36, -1) + 2 * 24;
        duration_var = toss_dice(24);
        if (duration_var > cycle_var) duration_var = cycle_var;
        discomfort_index_low = toss_multiple(5, 2, -1);
        discomfort_index_high = toss_multiple(5, 2, -1);
        if (discomfort_index_low > discomfort_index_high)
        {
            last_onset = discomfort_index_low;
            discomfort_index_low = discomfort_index_high;
            discomfort_index_high = last_onset;
        }
        last_onset = toss_dice(36);
        if (last_onset >= (int)rint((cycle_length + cycle_var) / 24.0))
            last_onset -= (int)rint((cycle_length + cycle_var) / 24.0);
        fprintf(stdout, "Cycle: %2.1f days (%d hrs) (+/- %d hrs) onset to onset\n",
                         cycle_length / 24.0, cycle_length, cycle_var);
        fprintf(stdout, "\tDuration: %2.1f days (%d hrs) (+/- %d hrs)\n",
                         duration / 24.0, duration, duration_var);
        if (discomfort_index_low != discomfort_index_high)
            fprintf(stdout, "\tDiscomfort index: %d-%d\n", discomfort_index_low, discomfort_index_high);
        else
            fprintf(stdout, "\tDiscomfort index: %d\n", discomfort_index_high);
        fprintf(stdout, "\tLast onset: %d days ago%s\n", last_onset,
                        ((int)rint(duration / 24.0) > last_onset) ? ", IT'S NOT FUNNY!" : "" );
    }
    while (prophecy_flag)
    {
        fprintf(stdout, "In accordance with the prophecy.\n");
        prophecy_flag--;
    }
    if (salute_flag)
    	fprintf(stdout, "Be seeing you!\n");
    return 0;
}
