/* Growing Pains
 *
 * Given: sex, adult height, and BMI, generates height, weight and BMI
 * for birthdays 2-20.
 * Copyright (c) 2005 Amelia A. Lewis
 *
 * See COPYING for license.  This work is released under the GPL,
 * according to the prophecy.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/param.h>
#include <math.h>

#include "instloc.h"

char *progname;
char height_string[MAXPATHLEN], bmi_string[MAXPATHLEN];

int adult_height = 0;
double adult_bmi = 0.0;

int poison_flag = -1;
int random_flag = 0;
int single_age = 0;

int hpercentm[] = { 0, 0, 49, 53, 58, 62, 65, 69, 72, 75, 79, 81, 84, 88, 93, 96, 98, 99, 99, 100, 100 };
int hpercentf[] = { 0, 0, 52, 57, 62, 66, 70, 74, 78, 81, 84, 88, 92, 96, 98, 99, 99, 100, 100, 100, 100 };

int bpercentm[] = { 0, 0, 72, 70, 68, 67, 67, 67, 69, 70, 72, 75, 77, 80, 83, 86, 89, 92, 95, 98, 100 };
int bpercentf[] = { 0, 0, 75, 72, 70, 70, 70, 70, 72, 75, 77, 80, 83, 86, 89, 91, 94, 96, 98, 99, 100 };

/*
 * usage
 *	Prints a usage statement.  Exits with return 1.
 */
void
usage(char *calledas)
{
    fprintf(stderr, "Usage:\n\t%s [-fhmrv] [-a age] [height] [bmi]\
\n\
\n      -a  Supply an age from 2-20, generate only that\
\n      -f  Female\
\n      -h  Show this usage description and exit\
\n      -m  Male\
\n      -r  Throw in some random variation\
\n      -v  Show version and exit\
\n\
\n      height: in inches (required) (5 feet is 60 inches), in the range 48-84\
\n      bmi: to one decimal place, please, in the range 13-40\
\n\
\n      Note: when selecting an age, you still must give adult height and bmi.\
\n", calledas);
    exit(1);
}

void
version(char *calledas)
{
    fprintf(stderr, "%s version %s\n", calledas, VERSION);
    exit(1);
}

int
toss_dice(int range)
{
    return ((int) ((1.0) * range * random() / (RAND_MAX + 1.0)) - (range / 2));
}

/*
 * getargs
 *	Parses the command line, setting flags.
 */
void
getargs(int ac, char *av[])
{
    int c;
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);

    while ((c = getopt(ac, av, "a:fhmrv")) != EOF)
        switch (c)
        {
            case 'a':
                if (*optarg)
                {
                    single_age = (int)strtoul(optarg, NULL, 10);
                    if ((single_age < 2) || (single_age > 20))
                        single_age = 0;
                }
                break;
            case 'f':
                poison_flag = 0;
                break;
            case 'm':
                poison_flag = 1;
                break;
            case 'r':
                random_flag = 1;
                break;
            case 'v':
                version(progname);
                break;
            case 'h':
            case '?':
            default:
                usage(progname);
                break;
        }
        
    if (poison_flag == -1)
        usage(progname);

    av += optind;
    if (*av)
        strcpy(height_string, *av);
    else usage(progname);
    
    if (*++av)
        strcpy(bmi_string, *av);
    else usage(progname);

    sscanf(height_string, "%d", &adult_height);
    if ( (adult_height < 48) || (adult_height > 84) )
        usage(progname);
    
    sscanf(bmi_string, "%lf", &adult_bmi);
    if ( (adult_bmi < 13.0) || (adult_bmi > 40.0) )
        usage(progname);
}

int
get_height(int percent)
{
    /* plus random variation? */
    int variance = 0;
    if (random_flag)
        variance = toss_dice(6);
    while ( (percent + variance) > 100)
        variance--;
    return (int)(adult_height * (percent + variance) / 100.0);
}

double
get_bmi(int percent)
{
    /* plus random variation? */
    int variance = 0;
    if (random_flag)
        variance = toss_dice(10);
    return adult_bmi * (percent + variance) / 100.0; 
}

int
get_weight(int height, double bmi)
{
    double weight = bmi * height * height / 703.0;
    return (int)rint(weight);
}

void
generate_descr(int year)
{
    static int last_height = 0;
    int hpercent = (poison_flag ? hpercentm[year] : hpercentf[year]);
    int bpercent = (poison_flag ? bpercentm[year] : bpercentf[year]);
    
    int height = get_height(hpercent);
    double bmi = get_bmi(bpercent);
    
    /* handle random growth spurts */
    if (height < last_height)
        height = last_height;
    last_height = height;
    
    fprintf(stdout, "%2d : %d'%2d\", %3d lbs. (BMI %3.1f)\n", year, height / 12, height % 12,
            get_weight(height, bmi), bmi);
}

/*
 * main
 *	If you need a comment for main, I suggest you try K&R
 */
int
main(int argc, char *argv[])
{
    int i;
    
    getargs(argc, argv);
    srandom(time(NULL) + getpid());

    if (single_age)
        generate_descr(single_age);
    else for (i = 2; i < 21; i++)
    {
        generate_descr(i);
    }

    return 0;
}
