/* riches
 *
 * Given a starting date, length+variation, duration+variation, discomfort
 * range, and number of instances, generates a menstrual cycle.
 *
 * rags to riches, you see.
 *
 * err, sorry.
 *
 * Copyright (c) 2005 Amelia A. Lewis
 *
 * See COPYING for license.  This work is released under the GPL,
 * according to the prophecy.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/param.h>
#include <math.h>

#include "instloc.h"

char *progname;

int year, month, day, hour, year2, month2, day2, hour2;
int julianstart, julianstop;

int len, lenv;

int dur, durv;

int painl, painu;

int nreps, counter = 0;
/*
 * usage
 *	Prints a usage statement.  Exits with return 1.
 */
void
usage(char *calledas)
{
    fprintf(stderr, "Usage:\n\t%s YYYY MM DD HH LLL Lv DD Dv Pl Pu N\
\n\
\n      YYYY year of first onset\
\n      MM   month of first onset\
\n      DD   day of first onset\
\n      HH   hour of first onset\
\n      LLL  average length, in hours\
\n      Lv   variation in length, in hours\
\n      DD   average duration, in hours\
\n      Dv   variation in duration, in hours\
\n      Pl   pain index lower bound\
\n      Pu   pain index upper bound\
\n      N    number of cycles to generate\
\n\
\n      Displays N lines of the form:\
\n      Onset: YYYY MM DD HH; Cessation: YYYY MM DD HH; Discomfort: P\
\n", calledas);
    exit(1);
}

void
version(char *calledas)
{
    fprintf(stderr, "%s version %s\n", calledas, VERSION);
    exit(1);
}

int
toss_dice(int upper_limit)
{
    return (int) ((upper_limit * 1.0) * random() / (RAND_MAX + 1.0));
}

/*
 * getargs
 *	Parses the command line, setting flags.
 */
void
getargs(int ac, char *av[])
{
    int c, is_leap;
    
    if (strrchr(av[0], '/'))
        progname = strdup(strrchr(av[0], '/'));
    else
        progname = strdup(av[0]);

    while ((c = getopt(ac, av, "hv")) != EOF)
        switch (c)
        {
            case 'v':
                version(progname);
                break;
            case 'h':
            case '?':
            default:
                usage(progname);
                break;
        }

    av += optind;
    if (*av)
        year = (int)strtoul(*av, NULL, 10);
    else usage(progname);
    if (*++av)
        month = (int)strtoul(*av, NULL, 10);
    else usage(progname);
    if (*++av)
        day = (int)strtoul(*av, NULL, 10);
    else usage(progname);
    if (*++av)
        hour = (int)strtoul(*av, NULL, 10);
    else usage(progname);
    if (*++av)
        len = (int)strtoul(*av, NULL, 10);
    else usage(progname);
    if (*++av)
        lenv = (int)strtoul(*av, NULL, 10);
    else usage(progname);
    if (*++av)
        dur = (int)strtoul(*av, NULL, 10);
    if (*++av)
        durv = (int)strtoul(*av, NULL, 10);
    else usage(progname);
    if (*++av)
        painl = (int)strtoul(*av, NULL, 10);
    else usage(progname);
    if (*++av)
        painu = (int)strtoul(*av, NULL, 10);
    else usage(progname);
    if (*++av)
        nreps = (int)strtoul(*av, NULL, 10);
    else usage(progname);

    year2 = year; 
    is_leap = year % 4 ? 0 : 1;   
    switch (month)
    {
        case 1:
            julianstart = day;
            break;
        case 2:
            julianstart = day + 31;
            break;
        case 3:
            julianstart = day + 59 + is_leap;
            break;
        case 4:
            julianstart = day + 90 + is_leap;
            break;
        case 5:
            julianstart = day + 120 + is_leap;
            break;
        case 6:
            julianstart = day + 151 + is_leap;
            break;
        case 7:
            julianstart = day + 181 + is_leap;
            break;
        case 8:
            julianstart = day + 212 + is_leap;
            break;
        case 9:
            julianstart = day + 243 + is_leap;
            break;
        case 10:
            julianstart = day + 273 + is_leap;
            break;
        case 11:
            julianstart = day + 304 + is_leap;
            break;
        case 12:
            julianstart = day + 334 + is_leap;
            break;
    }
/*fprintf(stderr, "%d repetitions\n", nreps);
fprintf(stderr, "Start %d %d %d %d\n", year, month, day, hour);
fprintf(stderr, "Length %d +/- %d\n", len, lenv);
fprintf(stderr, "Duration %d +/- %d\n", dur, durv);
fprintf(stderr, "Discomfort %d-%d\n", painl, painu);*/
}

int
get_julian_start(int incr_hours)
{
    julianstart = julianstart + (int)(incr_hours / 24);
    if ((incr_hours % 24) + hour > 23)
        julianstart++;
    if (!(year % 4) && (julianstart > 366))
    {
        year++;
        julianstart -= 366;
    }
    else if (julianstart > 365)
    {
        year++;
        julianstart -= 365;
    }
    return (incr_hours + hour) % 24;
}

int
get_julian_stop(int incr_hours)
{
    year2 = year;
    julianstop = julianstart + (int)(incr_hours / 24);
    if ((incr_hours % 24) + hour > 23)
        julianstop++;
    if (!(year % 4) && (julianstop > 366))
    {
        year2++;
        julianstop -= 366;
    }
    else if (julianstop > 365)
    {
        year2++;
        julianstop -= 365;
    }
    return (incr_hours + hour) % 24;
}

void
decode_julian(int is_duration)
{
    int *m, *d, *j, *y;
    if (is_duration)
    {
        j = &julianstop;
        m = &month2;
        d = &day2;
        y = &year2;
    }
    else
    {
        j = &julianstart;
        m = &month;
        d = &day;
        y = &year;
    }
    if (!(*y % 4))
    {
        switch (*j)
        {
            case 1 ... 31:
                *m = 1;
                *d = *j;
                break;
            case 32 ... 60:
                *m = 2;
                *d = *j - 31;
                break;
            case 61 ... 91:
                *m = 3;
                *d = *j - 60;
                break;
            case 92 ... 121:
                *m = 4;
                *d = *j - 91;
                break;
            case 122 ... 152:
                *m = 5;
                *d = *j - 121;
                break;
            case 153 ... 182:
                *m = 6;
                *d = *j - 152;
                break;
            case 183 ... 213:
                *m = 7;
                *d = *j - 182;
                break;
            case 214 ... 244:
                *m = 8;
                *d = *j - 213;
                break;
            case 245 ... 274:
                *m = 9;
                *d = *j - 244;
                break;
            case 275 ... 305:
                *m = 10;
                *d = *j - 274;
                break;
            case 306 ... 335:
                *m = 11;
                *d = *j - 305;
                break;
            case 336 ... 366:
                *m = 12;
                *d = *j - 335;
                break;
        }
    }
    else
    {
        switch (*j)
        {
            case 1 ... 31:
                *m = 1;
                *d = *j;
                break;
            case 32 ... 59:
                *m = 2;
                *d = *j - 31;
                break;
            case 60 ... 90:
                *m = 3;
                *d = *j - 59;
                break;
            case 91 ... 120:
                *m = 4;
                *d = *j - 90;
                break;
            case 121 ... 151:
                *m = 5;
                *d = *j - 120;
                break;
            case 152 ... 181:
                *m = 6;
                *d = *j - 151;
                break;
            case 182 ... 212:
                *m = 7;
                *d = *j - 181;
                break;
            case 213 ... 243:
                *m = 8;
                *d = *j - 212;
                break;
            case 244 ... 273:
                *m = 9;
                *d = *j - 243;
                break;
            case 274 ... 304:
                *m = 10;
                *d = *j - 273;
                break;
            case 305 ... 334:
                *m = 11;
                *d = *j - 304;
                break;
            case 335 ... 365:
                *m = 12;
                *d = *j - 334;
                break;
        }
    }
}

int
incr_hours(int is_duration)
{
    int positive = toss_dice(2);
    if (positive)
    {
        if (is_duration)
            return dur + toss_dice(durv + 1);
        else
            return len + toss_dice(lenv + 1);
    }
    else
    {
        if (is_duration)
            return dur - toss_dice(durv + 1);
        else
            return len - toss_dice(lenv + 1);
    }
}

/*
 * main
 *	If you need a comment for main, I suggest you try K&R
 */
int
main(int argc, char *argv[])
{
    int pain;
    srandom(time(NULL) + getpid());
    getargs(argc, argv);

    while (counter < nreps)
    {
        counter++;
        hour2 = get_julian_stop(incr_hours(1));
        decode_julian(1);
        if (painl == painu)
            pain = painl;
        else
            pain = painl + toss_dice(painu - painl + 1);
        fprintf(stdout, "Onset: %d %d %d %d; Cessation: %d %d %d %d; Discomfort: %d\n", 
                         year, month, day, hour, year2, month2, day2, hour2, pain);
        hour = get_julian_start(incr_hours(0));
        decode_julian(0);
    }

    /*    Onset: YYYY MM DD HH - YYYY MM DD HH, discomfort P-P */
    return 0;
}
